#include <OgreNoMemoryMacros.h>
#include <CEGUI/CEGUI.h>
#include <../CEGUIRenderer/include/OgreCEGUIRenderer.h>
#include <../CEGUIRenderer/include/OgreCEGUIResourceProvider.h>
#include <../CEGUIRenderer/include/OgreCEGUITexture.h>
#include <OgreMemoryMacros.h>
#include <OgreIteratorWrappers.h>
#include <Ogre.h>
#include <direct.h> // for getcwd
//#include "OgreReferenceAppLayer.h"
//#include "OgreRefAppWorld.h"
#include "TestCullingTerrainApplication.h"
#include "TerrainFrameListener.h"
//#include "OgreVisibilityOctreeSceneManager.h"

#include "IVReader.h"

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

const static bool USE_STATIC_GEOMETRY = false;
bool TestCullingTerrainApplication::msShowHillyTerrain = false;
//bool TestCullingTerrainApplication::msShowHillyTerrain = true;

/******************************************************/
/*             EntityState implementation             */
/******************************************************/


Vector3 EntityState::msMinPos = Vector3::ZERO;
Vector3 EntityState::msMaxPos = Vector3::ZERO;

EntityState::EntityState(Entity *ent, State entityState, Real speed):
mEntity(ent), mState(entityState), mAnimationSpeed(speed)
{
	switch(entityState)
	{
	case MOVING:
		mAnimationState = mEntity->getAnimationState("Walk");
		break;
	case WAITING:
		mAnimationState = mEntity->getAnimationState("Idle");
		break;
	case STOP:
		mAnimationState = NULL;
		break;
	default:
		break;
	}
	// enable animation state
	if (mAnimationState)
	{
		mAnimationState->setLoop(true);
		mAnimationState->setEnabled(true);
	}

	mTimeElapsed = Math::RangeRandom(1, 5);
}
//-----------------------------------------------------------------------
EntityState::~EntityState()
{
	mAnimationState = NULL;
	mEntity = NULL;
}
//-----------------------------------------------------------------------
Entity *EntityState::GetEntity()
{
	return mEntity;
}
//-----------------------------------------------------------------------
EntityState::State EntityState::GetState()
{
	return mState;
}
//-----------------------------------------------------------------------
void EntityState::update(Real timeSinceLastFrame)
{
	mTimeElapsed -= timeSinceLastFrame * mAnimationSpeed;

	if (!mEntity || !mAnimationState)
		return;
	
	if (mState == MOVING) // toggle between moving (longer) and waiting (short)
	{
		SceneNode *parent = mEntity->getParentSceneNode();
		
		if (!parent)
			return;

		if (mTimeElapsed <= 0) // toggle animation state
		{
			if (mAnimationState->getAnimationName() == "Idle")
			{
				SetAnimationState("Walk", true);
				
				mTimeElapsed = walk_duration; // walk for mTimeElapsed units

				// choose random rotation
				Radian rnd = Radian(Math::UnitRandom() * Math::PI * rotate_factor);

				//mEntity->getParentSceneNode()->rotate();
				parent->yaw(rnd);						
			}
			else 
			{
				SetAnimationState("Idle", true);
				mTimeElapsed = wait_duration; // wait for mTimeElapsed seconds
			}
		}

		if (mAnimationState->getAnimationName() == "Walk") // move forward
		{
			// store old position, just in case we get out of bounds
			Vector3 oldPos = parent->getPosition();
			parent->translate(parent->getLocalAxes(), Vector3(move_factor * mAnimationSpeed, 0, 0));

			// HACK: if out of bounds => reset to old position and set animationstate to idle
			if (OutOfBounds(parent)) 
			{
				parent->setPosition(oldPos);
				SetAnimationState("Idle", true);
				
				mTimeElapsed = wait_duration;
			}
		}
	}
		
	// add time to drive animation
	mAnimationState->addTime(timeSinceLastFrame * mAnimationSpeed);
}
//-----------------------------------------------------------------------
void EntityState::SetAnimationState(String stateStr, bool loop)
{
	if (!mEntity)
		return;

	mAnimationState = mEntity->getAnimationState(stateStr);
	mAnimationState->setLoop(loop);
	mAnimationState->setEnabled(true);
}
//-----------------------------------------------------------------------
bool EntityState::OutOfBounds(SceneNode *node)
{
	Vector3 pos = node->getPosition();

	if ((pos > msMinPos) && (pos < msMaxPos))
		return false;

	return true;
}



/*********************************************************/
/*      TestCullingTerrainApplication implementation     */
/*********************************************************/


TestCullingTerrainApplication::TestCullingTerrainApplication():
mTerrainContentGenerator(NULL), 
mRayQueryExecutor(NULL),
mIVReader(NULL)
{
}
//-----------------------------------------------------------------------
TestCullingTerrainApplication::~TestCullingTerrainApplication()
{
	OGRE_DELETE(mTerrainContentGenerator);
	OGRE_DELETE(mRayQueryExecutor);
	OGRE_DELETE(mIVReader);

	deleteEntityStates();	
}
//-----------------------------------------------------------------------
void TestCullingTerrainApplication::deleteEntityStates()
{
	for (int i = 0; i < (int)mEntityStates.size(); ++i)
	{
		OGRE_DELETE(mEntityStates[i]);
	}

	mEntityStates.clear();
}
//-----------------------------------------------------------------------
void TestCullingTerrainApplication::createCamera()
{
	// create the camera
	mCamera = mSceneMgr->createCamera("PlayerCam");
	
	/** set a nice viewpoint
	*	we use a camera node here and apply all transformations on it instead
	*	of applying all transformations directly to the camera
	*	because then the camera is displayed correctly in the visualization
	*/
	mCamNode = mSceneMgr->getRootSceneNode()->
		createChildSceneNode("CamNode1", Vector3(707, 5000, 528));
	
	mCamNode->setOrientation(Quaternion(-0.3486, 0.0122, 0.9365, 0.0329));
	mCamNode->attachObject(mCamera);
	
	//-- create visualization camera
	mVizCamera = mSceneMgr->createCamera("VizCam");
	mVizCamera->setPosition(mCamNode->getPosition());

	mVizCamera->setNearClipDistance(1);
	mCamera->setNearClipDistance(1);

	// infinite far plane?
	if (mRoot->getRenderSystem()->getCapabilities()->hasCapability(RSC_INFINITE_FAR_PLANE))
	{
		mVizCamera->setFarClipDistance(0);
		mCamera->setFarClipDistance(0);
	}
	else
	{
		 mVizCamera->setFarClipDistance(20000);
		 mCamera->setFarClipDistance(20000);
	}	
}

//-----------------------------------------------------------------------
bool TestCullingTerrainApplication::setup()
{
	bool carryOn = ExampleApplication::setup();

	if (carryOn)
		createRenderTargetListener();

	return carryOn;
}
//-----------------------------------------------------------------------
void TestCullingTerrainApplication::createRenderTargetListener()
{
	mWindow->addListener(new VisualizationRenderTargetListener(mSceneMgr));
}
//-----------------------------------------------------------------------
void TestCullingTerrainApplication::LoadWorldGeometry()
{
	mIVReader = new IVReader();

	// _MAX_PATH is the maximum length allowed for a path
	char currentPath[_MAX_PATH];
	// use the function to get the path
	getcwd(currentPath, _MAX_PATH);
	

	// $$matt TODO; use relatrive path
	std::stringstream d;
	d << "current path: " << currentPath;
	Ogre::LogManager::getSingleton().logMessage(d.str());

	String path = "E:/svn/gametools/OGRE/trunk/resources/";
	

/*	SceneNode *node1 = mSceneMgr->getRootSceneNode()->createChildSceneNode("Node1");
		
	SceneNode *node2 = mSceneMgr->getRootSceneNode()->createChildSceneNode("Node2");
	SceneNode *node3 = mSceneMgr->getRootSceneNode()->createChildSceneNode("Node3");
	SceneNode *node4 = mSceneMgr->getRootSceneNode()->createChildSceneNode("Node4");
	SceneNode *node5 = mSceneMgr->getRootSceneNode()->createChildSceneNode("Node5");
	SceneNode *node6 = mSceneMgr->getRootSceneNode()->createChildSceneNode("Node6");
	SceneNode *node7 = mSceneMgr->getRootSceneNode()->createChildSceneNode("Node7");



	node2->translate(Vector3(-400, 0, 750));
	node3->scale(Vector3(100, 100 ,100));
	node3->pitch(Radian(3 * Math::HALF_PI));
	node4->translate(Vector3(-100, 0, 0));
	node5->translate(Vector3(0, 300, 0));
	
	node6->scale(Vector3(100, 100 ,100));
	node6->pitch(Radian(3 * Math::HALF_PI));
	node6->roll(Radian(Math::HALF_PI));
	node6->translate(Vector3(0, -100, 0));

	node7->scale(Vector3(100, 100 ,100)); 
	node7->pitch(Radian(3 * Math::HALF_PI));
	node7->roll(Radian(Math::HALF_PI));
	node7->translate(Vector3(100, 0, 0));
	
	*/

	Log *log = LogManager::getSingleton().createLog("IVLog.log");
	mIVReader->setLog(log);



#if 0
	//String fileName = path + "media\\simple_geom.wrl";
	String fileName = path + "media\\city.wrl";
    //String fileName = path + "media\\house_test1.iv";
	mCamNode->setPosition(Vector3(-6, 0.5, 10));
	//mCamNode->setPosition(Vector3(412, 7.4, -667));
	mCamNode->setOrientation(Quaternion(-0.95, 0, 0.3, 0));

	if (0 && mIVReader->loadFile(fileName.c_str()))
	{
		mIVReader->buildTree(mSceneMgr, node1);
	}
	
	String fileName = path + "media\\house_test1.iv";

	if (0 && mIVReader->loadFile(fileName.c_str()))
	{
		mIVReader->buildTree(mSceneMgr, node1);
		mIVReader->buildTree(mSceneMgr, node2);
	}

	if (0 && mIVReader->loadFile("media\\anhk.wrl"))
	{
		mIVReader->buildTree(mSceneMgr, node3);
	}
	if (0 && mIVReader->loadFile("media\\ant.wrl"))
	{
		mIVReader->buildTree(mSceneMgr, node4);
	}
	if (0 && mIVReader->loadFile("media\\zepplin0.wrl"))
	{
		mIVReader->buildTree(mSceneMgr, node5);
	}
	if (0 && mIVReader->loadFile("media\\warbird.wrl"))
	{
		mIVReader->buildTree(mSceneMgr, node6);
	}
	if (0 && mIVReader->loadFile("media\\z_wing.wrl"))
	{
		mIVReader->buildTree(mSceneMgr, node7);
	}

#endif
	
#if 1
	//-- load vienna model

	SceneNode *roofsNode = NULL;
	SceneNode *viennaNode = NULL;
	SceneNode *roadsNode = NULL;
	SceneNode *planeNode = NULL;
	SceneNode *wallsNode = NULL;

	viennaNode = mSceneMgr->getRootSceneNode()->createChildSceneNode("Vienna");

	mCamNode->setPosition(Vector3(830, 300, -540));
	mCamNode->setOrientation(Quaternion(-0.3486, 0.0122, 0.9365, 0.0329));
	
	//viennaNode->translate(Vector3(-300, -300, 0));

	String fileName = path + "media\\city1500_flat_1.iv";

	if (0 && mIVReader->loadFile(fileName.c_str()))
	{
		wallsNode = viennaNode->createChildSceneNode("Walls");
		mIVReader->buildTree(mSceneMgr, wallsNode);
	}


	fileName = path + "media\\roofs_1500.iv";
	if (1 && mIVReader->loadFile(fileName.c_str()))
	{
		roofsNode = viennaNode->createChildSceneNode("Roofs");
		mIVReader->buildTree(mSceneMgr, roofsNode);	
	}
    
	fileName = path + "media\\CityRoads60.iv";
	if (0 && mIVReader->loadFile(fileName.c_str()))
	{
		roadsNode = viennaNode->createChildSceneNode("Roads");
		mIVReader->buildTree(mSceneMgr, roadsNode);	
	}

	fileName = path + "media\\CityPlane60.iv";
    if (0 && mIVReader->loadFile(fileName.c_str()))
	{
		planeNode = viennaNode->createChildSceneNode("Plane");
		mIVReader->buildTree(mSceneMgr, planeNode);		
	}
	
#endif

	mIVReader->collapse();

	//-- bake into static geometry
	if (USE_STATIC_GEOMETRY)
	{
		BakeSceneIntoStaticGeometry("staticVienna", "Vienna");
	}
}
//--------------------------------------------------------
void TestCullingTerrainApplication::BakeSceneIntoStaticGeometry(const String &staticGeomName, 
																const String &nodeName)
{
#if OGRE_103
	// note: different static geom for roofs, roads, ..., becazse they have same material
	StaticGeometry *staticGeom = mSceneMgr->createStaticGeometry(staticGeomName);

	// note: looping over entities here. why does scene node not work?
	SceneManager::EntityIterator it = mSceneMgr->getEntityIterator();
	while (it.hasMoreElements())
	{
		Entity *ent = it.getNext();
		ent->setVisible(false);
		staticGeom->addEntity(ent, ent->getParentSceneNode()->getPosition());
	}

	staticGeom->setRegionDimensions(Vector3(100,100,100));
	staticGeom->build();

	// cleanup node
	//wallsNode->detachAllObjects();
	
	//roofsNode->detachAllObjects();
//roadsNode->detachAllObjects();
//	planeNode->detachAllObjects();
	
	//viennaNode->removeChild("Walls");
	mSceneMgr->destroySceneNode(nodeName);
#endif
}
//--------------------------------------------------------
void TestCullingTerrainApplication::createScene()
{
	if (!msShowHillyTerrain)
		LoadWorldGeometry();

	// Set ambient light
	mAmbientLight = ColourValue(0.5, 0.5, 0.5);
	mSceneMgr->setAmbientLight(mAmbientLight);
	
	//-- create light
	mSunLight = mSceneMgr->createLight("SunLight");
	mSunLight->setType(Light::LT_DIRECTIONAL);
	//mSunLight->setType(Light::LT_SPOTLIGHT);
	//mSunLight->setSpotlightRange(Degree(30), Degree(50));

    mSunLight->setPosition(707, 2000, 500);
	mSunLight->setCastShadows(true);

	// set light angle not too small over the surface, otherwise shadows textures will be broken
	Vector3 dir(0.5, 1, 0.5);
	dir.normalise();
	mSunLight->setDirection(dir);
	//mSunLight->setDirection(Vector3::NEGATIVE_UNIT_Y);
	mSunLight->setDiffuseColour(1, 1, 1);
	mSunLight->setSpecularColour(1, 1, 1);

	// -- Fog
	// NB it's VERY important to set this before calling setWorldGeometry 
	// because the vertex program picked will be different
	ColourValue fadeColour(0.93, 0.86, 0.76);
	mWindow->getViewport(0)->setBackgroundColour(fadeColour);
	//mSceneMgr->setFog( FOG_LINEAR, fadeColour, .001, 500, 1000);
	
	// Create a skybox
	mSceneMgr->setSkyBox(true, "Examples/CloudyNoonSkyBox", 5000, true);
	
	if (msShowHillyTerrain)
	{
		std::string terrain_cfg("terrain.cfg");
#if OGRE_PLATFORM == OGRE_PLATFORM_APPLE
		terrain_cfg = mResourcePath + terrain_cfg;
#endif
		mSceneMgr->setWorldGeometry(terrain_cfg);
	}
	

	//-- CEGUI setup
	setupGui();

	// occluder plane to test visibility
	if (0)
	{
		Plane plane; plane.normal = Vector3::UNIT_Y;	plane.d = -60;
		MeshManager::getSingleton().createPlane("Myplane",
			ResourceGroupManager::DEFAULT_RESOURCE_GROUP_NAME, plane,
			5000,5000,100,100,true,1,5,5,Vector3::UNIT_Z);
		Entity* pPlaneEnt = mSceneMgr->createEntity( "plane", "Myplane" );
		pPlaneEnt->setMaterialName("Examples/Rockwall");
		pPlaneEnt->setCastShadows(true);
		mSceneMgr->getRootSceneNode()->createChildSceneNode()->attachObject(pPlaneEnt);
	}

	// Warning: In GL since we can't go higher than the window res
	mSceneMgr->setShadowTextureSettings(1024, 2);

	mSceneMgr->setShadowColour(ColourValue(0.5, 0.5, 0.5));
   

	//-- terrain content setup

	// HACK: necessary to call once before the content creation for
	// terrain initialisation
	mSceneMgr->_renderScene(mCamera, mWindow->getViewport(0), true);

	// ray query executor: needed to clamp to terrain
	mRayQueryExecutor = new RayQueryExecutor(mSceneMgr);

	mTerrainMinPos = EntityState::msMinPos = Vector3(0, 0, 0);
	mTerrainMaxPos = EntityState::msMaxPos = Vector3(5000, 5000, 5000);

	mTerrainContentGenerator = new TerrainContentGenerator(mSceneMgr);
	
	
	// if no objects in file, we generate new objects
	if (msShowHillyTerrain && !mTerrainContentGenerator->LoadObjects("objects.out"))
	{
		// the objects are generated randomly distributed over the terrain
		if (1) generateScene(500, 0); // create robots
		if (0) generateScene(100, 1); // create trees
		if (0) generateScene(100, 2); // create ninjas
	}

	// HACK: load the view cells from the disc
	//(dynamic_cast<VisibilityOctreeSceneManager *>(mSceneMgr))->LoadViewCells("");

	//std::string filename = "vienna_simple-21-04-avs2-viewCells.xml";
	//std::string filename = "vienna_simple-21-04-kdcasm-viewCells.xml";
	std::string filename = "vienna_simple-21-04-avs-viewCells.xml";

	mSceneMgr->setOption("LoadViewCells", filename.c_str());
}
//-----------------------------------------------------------------------
void  TestCullingTerrainApplication::generateScene(int num, int objectType)
{
	float val = TerrainFrameListener::msObjectScales[objectType];
	Vector3 scale(val, val, val); 
	const float maxHeight = 75;
	
	// In order to provide much occlusion,
	// height is restricted to maxHeight => no objects are created on peaks
	mTerrainContentGenerator->SetMinPos(Vector3(mTerrainMinPos));
	mTerrainContentGenerator->SetMaxPos(Vector3(mTerrainMaxPos.x, maxHeight, mTerrainMaxPos.z));
	
	//std::stringstream d; d << "objscale: " << scale[0];
	//Ogre::LogManager::getSingleton().logMessage(d.str());
	
	mTerrainContentGenerator->SetScale(scale);
	mTerrainContentGenerator->SetOffset(TerrainFrameListener::msObjectTerrainOffsets[objectType]);
	mTerrainContentGenerator->GenerateScene(num, TerrainFrameListener::msObjectCaptions[objectType]);

	if (objectType != 0) // from our objects, only robot has animation phases
		return;

	EntityList *entList = mTerrainContentGenerator->GetGeneratedEntities();

	//-- add animation state for new robots (located at the end of the list)
	for (int i = (int)entList->size() - num; i < (int)entList->size(); ++i)
	{
		mEntityStates.push_back(new EntityState((*entList)[i], 
			EntityState::WAITING, Math::RangeRandom(0.5, 1.5)));
	}

	// release limitations on height => it is possible for the user to put single 
	// objects on peaks of the terrain (will be only few, not relevant for occlusion)
	mTerrainContentGenerator->SetMaxPos(mTerrainMaxPos);
}
//-----------------------------------------------------------------------
void TestCullingTerrainApplication::updateAnimations(Real timeSinceLastFrame)
{
	for (int i = 0; i < (int)mEntityStates.size(); ++i)
	{
		SceneNode *sn = mEntityStates[i]->GetEntity()->getParentSceneNode();

		mEntityStates[i]->update(timeSinceLastFrame);

		if (mEntityStates[i]->GetState() == EntityState::MOVING)
		{
			Clamp2Terrain(sn, 0); //sn->setNodeVisible(false);
		}
	}
}
//-----------------------------------------------------------------------
EntityStates  &TestCullingTerrainApplication::getEntityStates()
{
	return mEntityStates;
}
//-----------------------------------------------------------------------
void TestCullingTerrainApplication::setupGui()
{
#if OGRE103
	 mGUIRenderer = new CEGUI::OgreCEGUIRenderer(mWindow, Ogre::RENDER_QUEUE_OVERLAY, 
		 false, 3000, ST_EXTERIOR_CLOSE);
#else
	  mGUIRenderer = new CEGUI::OgreCEGUIRenderer(mWindow, Ogre::RENDER_QUEUE_OVERLAY, 
		 false, 3000, mSceneMgr);
#endif
     mGUISystem = new CEGUI::System(mGUIRenderer);

	 // Mouse
     CEGUI::SchemeManager::getSingleton().loadScheme((CEGUI::utf8*)"TaharezLook.scheme");
     CEGUI::MouseCursor::getSingleton().setImage("TaharezLook", "MouseArrow");
	 mGUISystem->setDefaultMouseCursor((CEGUI::utf8*)"TaharezLook", 
									   (CEGUI::utf8*)"MouseArrow");

	 CEGUI::MouseCursor::getSingleton().hide();
}
//-----------------------------------------------------------------------
void TestCullingTerrainApplication::createFrameListener()
{
	mTerrainFrameListener = new TerrainFrameListener(mWindow, mCamera, mSceneMgr, 
		mGUIRenderer, mTerrainContentGenerator, mVizCamera, mCamNode, mSunLight, this);
	
	mRoot->addFrameListener(mTerrainFrameListener);
}
//-----------------------------------------------------------------------
void TestCullingTerrainApplication::chooseSceneManager()
{
#ifdef OGRE_103
	if (msShowHillyTerrain)
	{
		// Terrain scene manager
		mSceneMgr = mRoot->getSceneManager(ST_EXTERIOR_CLOSE);
	}
	else
	{	// octree scene manager
		mSceneMgr = mRoot->getSceneManager(ST_GENERIC);
	}
#else
	if (msShowHillyTerrain)
	{
		//mSceneMgr = mRoot->createSceneManager("VisibilityTerrainSceneManager");
		mSceneMgr = mRoot->createSceneManager("TerrainSceneManager");
	}
	else
	{	
		mSceneMgr = mRoot->createSceneManager("VisibilityOctreeSceneManager");
		//mSceneMgr = mRoot->createSceneManager("OctreeSceneManager");
	}

#endif
}
//-----------------------------------------------------------------------
bool TestCullingTerrainApplication::Clamp2Terrain(SceneNode *node, int terrainOffs)
{
	// clamp scene node to terrain
	Vector3 pos = node->getPosition();
	Vector3 queryResult;

	if (mRayQueryExecutor->executeRayQuery(&queryResult, 
			Vector3(pos.x, MAX_HEIGHT, pos.z), Vector3::NEGATIVE_UNIT_Y))
	{
        node->setPosition(pos.x, queryResult.y + terrainOffs, pos.z);
		return true;
	}

	return false;
}

//-----------------------------------------------------------------------
bool TestCullingTerrainApplication::Clamp2FloorPlane()
{
    // clamp to floor plane
	RaySceneQuery *raySceneQuery = mSceneMgr->createRayQuery(
		Ray(mCamNode->getPosition(), Vector3::NEGATIVE_UNIT_Y));

	RaySceneQueryResult& qryResult = raySceneQuery->execute();
    
	RaySceneQueryResult::iterator rit = qryResult.begin();
 
	while (rit != qryResult.end() && rit->movable)
	{
		if (rit->movable->getName() != "PlayerCam")
		{
			mCamNode->setPosition(mCamNode->getPosition().x,
				rit->movable->getWorldBoundingBox().getCenter().y + 2, 
				mCamNode->getPosition().z);
	
			/*std::stringstream d;
			d << "World: " << it->movable->getWorldBoundingBox().getCenter().y <<
			", Object: " << it->movable->getBoundingBox().getCenter().y <<
			", Camera: " << mCamera->getDerivedPosition();

			LogManager::getSingleton().logMessage(d.str());*/

			return true;
		}

		++ rit;
	}
    
	OGRE_DELETE(raySceneQuery);
	return false;
}



/**********************************************************************/
/*           VisualizationRenderTargetListener implementation         */
/**********************************************************************/


//-----------------------------------------------------------------------
VisualizationRenderTargetListener::VisualizationRenderTargetListener(SceneManager *sceneMgr)
:RenderTargetListener(), mSceneMgr(sceneMgr)
{
}
//-----------------------------------------------------------------------
void VisualizationRenderTargetListener::preViewportUpdate(const RenderTargetViewportEvent &evt)
{
	// visualization viewport
	const bool showViz = evt.source->getZOrder() == VIZ_VIEWPORT_Z_ORDER;
	const bool nShowViz = !showViz;

	mSavedShadowTechnique = mSceneMgr->getShadowTechnique();
	mSavedAmbientLight = mSceneMgr->getAmbientLight();

	// -- ambient light must be full for visualization, shadows disabled
    if (showViz)
	{
		mSceneMgr->setAmbientLight(ColourValue(1, 1, 1));
		mSceneMgr->setShadowTechnique(SHADOWTYPE_NONE);
	}
	
    mSceneMgr->setOption("PrepareVisualization", &showViz);
	mSceneMgr->setOption("SkyBoxEnabled", &nShowViz);
	//mSceneMgr->setOption("SkyPlaneEnabled", &showViz);
	
	RenderTargetListener::preViewportUpdate(evt);
}
//-----------------------------------------------------------------------
void VisualizationRenderTargetListener::postRenderTargetUpdate(const RenderTargetEvent &evt)
{
	// reset values
	mSceneMgr->setShadowTechnique(mSavedShadowTechnique);
	mSceneMgr->setAmbientLight(mSavedAmbientLight);
	
	RenderTargetListener::postRenderTargetUpdate(evt);
}
//-----------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR strCmdLine, INT )
{
    // Create application object
    TestCullingTerrainApplication app;

	try 
	{
        app.go();
    } 
	catch( Ogre::Exception& e ) 
	{
        MessageBox( NULL, e.getFullDescription().c_str(), 
			"An exception has occured!", MB_OK | MB_ICONERROR | MB_TASKMODAL);
    }	

    return 0;
}
